"""Dotted dictionary for consistent interface.

Consider moving to Corallium, but I don't have any uses for it yet.

"""

from beartype import beartype
from beartype.typing import Any, Dict, Union
from box import Box

DdictType = Union[Dict[str, Any], Box]
"""Return type from `ddict()`."""


@beartype
def ddict(**kwargs: Dict[str, Any]) -> DdictType:
    """Return a dotted dictionary that can also be accessed normally.

    - Currently uses `python-box`
    - Could consider `cleverdict` which had updates as recently as 2022
    - There are numerous other variations that haven't been updated since 2020, such as `munch`, `bunch`, `ddict`

    Args:
        **kwargs: keyword arguments formatted into dictionary

    Returns:
        DdictType: dotted dictionary

    """
    return Box(kwargs)



"""Invoke Helpers."""

import platform
from contextlib import suppress
from functools import lru_cache
from os import environ
from pathlib import Path

from beartype import beartype
from beartype.typing import Any, Optional
from corallium.file_helpers import COPIER_ANSWERS, read_yaml_file
from invoke.context import Context
from invoke.runners import Result

# ----------------------------------------------------------------------------------------------------------------------
# General Invoke


@lru_cache(maxsize=1)
def use_pty() -> bool:
    """Returns False on Windows and some CI environments."""
    if platform.system() == 'Windows':
        return False
    return not environ.get('GITHUB_ACTION')


@beartype
def run(ctx: Context, *run_args: Any, **run_kwargs: Any) -> Optional[Result]:
    """Wrap invoke.run to run within the `working_dir`."""
    working_dir = '.'
    with suppress(AttributeError):
        working_dir = ctx.config.gto.working_dir

    with ctx.cd(working_dir):
        return ctx.run(*run_args, **run_kwargs)


# ----------------------------------------------------------------------------------------------------------------------
# Invoke Task Helpers


@lru_cache(maxsize=1)
def get_project_path() -> Path:
    """Retrieve the `cwd`."""
    return Path.cwd()


@beartype
def get_doc_subdir(path_project: Optional[Path] = None) -> Path:
    """Retrieve the documentation directory from the copier answer file.

    Args:
        path_project: Path to the project directory with contains `.copier-answers.yml`

    Returns:
        Path: to the source documentation directory

    """
    path_copier = (path_project or get_project_path()) / COPIER_ANSWERS
    doc_dir = read_yaml_file(path_copier).get('doc_dir', 'docs')
    return path_copier.parent / doc_dir / 'docs'  # type: ignore[no-any-return]



"""Nox CLI."""

from invoke.context import Context

from ..cli import task
from ..invoke_helpers import run
from .executable_utils import python_dir


@task(
    default=True,
    help={
        'session': 'Optional session to run',
    },
)
def noxfile(ctx: Context, *, session: str = '') -> None:
    """Run nox from the local noxfile."""
    cli_args = ['--session', session] if session else []
    run(ctx, f'{python_dir()}/nox --error-on-missing-interpreters {" ".join(cli_args)}')



"""Extend Invoke for Calcipy."""

import sys
from base64 import b64encode
from functools import wraps
from pathlib import Path
from types import ModuleType

from beartype import beartype
from beartype.typing import Any, Callable, Dict, List, Optional, Union
from invoke.collection import Collection as InvokeCollection  # noqa: TID251
from invoke.config import Config, merge_dicts
from invoke.program import Program

from .invoke_helpers import use_pty
from .tasks._invoke import TASK_ARGS_ATTR, TASK_KWARGS_ATTR, Collection, GlobalTaskOptions


class _CalcipyProgram(Program):
    """Customized version of Invoke's `Program`."""

    def print_help(self) -> None:  # pragma: no cover
        """Extend print_help with calcipy-specific global configuration.

        https://github.com/pyinvoke/invoke/blob/0bcee75e4a26ad33b13831719c00340ca12af2f0/invoke/program.py#L657-L667

        """
        super().print_help()
        print('Global Task Options:')  # noqa: T201
        print()  # noqa: T201
        self.print_columns([
            ('*file_args', 'List of Paths available globally to all tasks. Will resolve paths with working_dir'),
            ('--keep-going', 'Continue running tasks even on failure'),
            ('--working_dir=STRING', 'Set the cwd for the program. Example: "../run --working-dir .. lint test"'),
            ('-v,-vv,-vvv', 'Globally configure logger verbosity (-vvv for most verbose)'),
        ])
        print()  # noqa: T201


class CalcipyConfig(Config):
    """Opinionated Config with better defaults."""

    @staticmethod
    def global_defaults() -> Dict:  # type: ignore[type-arg]  # pragma: no cover
        """Override the global defaults."""
        invoke_defaults = Config.global_defaults()
        calcipy_defaults = {
            'run': {
                'echo': True,
                'echo_format': '\033[2;3;37mRunning: {command}\033[0m',
                'pty': use_pty(),
            },
        }
        return merge_dicts(invoke_defaults, calcipy_defaults)


@beartype
def start_program(  # noqa: CAC001
    pkg_name: str,
    pkg_version: str,
    module: Optional[ModuleType] = None,
    collection: Optional[Union[Collection, InvokeCollection]] = None,
) -> None:  # pragma: no cover
    """Run the customized Invoke Program.

    FYI: recommendation is to extend the `core_args` method, but this won't parse positional arguments:
    https://docs.pyinvoke.org/en/stable/concepts/library.html#modifying-core-parser-arguments

    """
    # Manipulate 'sys.argv' to hide arguments that invoke can't parse
    _gto = GlobalTaskOptions()
    sys_argv: List[str] = sys.argv[:1]
    last_argv = ''
    for argv in sys.argv[1:]:
        if not last_argv.startswith('-') and Path(argv).is_file():
            _gto.file_args.append(Path(argv))
        # Check for CLI flags
        elif argv in {'-v', '-vv', '-vvv', '--verbose'}:
            _gto.verbose = argv.count('v')
        elif argv == '--keep-going':
            _gto.keep_going = True
        # Check for CLI arguments with values
        elif last_argv == '--working-dir':
            _gto.working_dir = Path(argv).resolve()
        elif argv != '--working-dir':
            sys_argv.append(argv)
        last_argv = argv
    _gto.file_args = [
        _f if _f.is_absolute() else Path.cwd() / _f
        for _f in _gto.file_args
    ]
    sys.argv = sys_argv

    class _CalcipyConfig(CalcipyConfig):

        gto: GlobalTaskOptions = _gto

    if module and collection:
        raise ValueError('Only one of collection or module can be specified')

    _CalcipyProgram(
        name=pkg_name,
        version=pkg_version,
        namespace=Collection.from_module(module) if module else collection,
        config_class=_CalcipyConfig,
    ).run()


def task(*dec_args: Any, **dec_kwargs: Any) -> Callable:  # type: ignore[type-arg]
    """Marks wrapped callable object as a valid Invoke task."""
    def wrapper(func: Any) -> Callable:  # type: ignore[type-arg]
        # Attach arguments for Task
        setattr(func, TASK_ARGS_ATTR, dec_args)
        setattr(func, TASK_KWARGS_ATTR, dec_kwargs)
        # Attach public attributes from invoke that are expected
        func.help = dec_kwargs.pop('help', {})

        def _with_kwargs(**extra_kwargs: Any) -> Callable:  # type: ignore[type-arg] # nosem
            """Support building partial tasks."""
            if extra_kwargs:
                # Set a unique name when 'extra_kwargs' was provided
                # https://github.com/pyinvoke/invoke/blob/07b836f2663bb073a7bcef3d6c454e1dc6b867ae/invoke/tasks.py#L81-L104
                encoded = b64encode(str(extra_kwargs).encode())
                func.__name__ = f'{func.__name__}_{encoded.decode().rstrip("=")}'

            @wraps(func)  # nosem
            def _with_kwargs_inner(*args: Any, **kwargs: Any) -> Any:
                return func(*args, **kwargs, **extra_kwargs)
            return _with_kwargs_inner

        func.with_kwargs = _with_kwargs

        @wraps(func)  # nosem
        def _inner(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        return _inner

    # Handle the case when the decorator is called without arguments
    if (
        len(dec_args) == 1
        and callable(dec_args[0])
        and not hasattr(dec_args[0], TASK_ARGS_ATTR)
    ):
        return wrapper(dec_args[0])

    return wrapper



"""Code Tag Collector CLI."""

from pathlib import Path

from beartype.typing import Optional
from invoke.context import Context

from ..cli import task
from ..code_tag_collector import write_code_tag_file
from ..file_search import find_project_files
from ..invoke_helpers import get_doc_subdir
from .defaults import from_ctx


@task(
    default=True,
    help={
        'base_dir': 'Working Directory',
        'doc_sub_dir': 'Subdirectory for output of the code tag summary file',
        'filename': 'Code Tag Summary Filename',
        'tag_order': 'Ordered list of code tags to locate (Comma-separated)',
        'regex': 'Custom Code Tag Regex. Must contain "{tag}"',
        'ignore_patterns': 'Glob patterns to ignore files and directories when searching (Comma-separated)',
    },
)
def collect_code_tags(  # noqa: PLR0913,PLR0917
    ctx: Context,
    base_dir: str = '.',
    doc_sub_dir: str = '',
    filename: Optional[str] = None,
    tag_order: str = '',
    regex: str = '',
    ignore_patterns: str = '',
) -> None:
    """Create a `CODE_TAG_SUMMARY.md` with a table for TODO- and FIXME-style code comments."""
    pth_base_dir = Path(base_dir).resolve()
    pth_docs = pth_base_dir / doc_sub_dir if doc_sub_dir else get_doc_subdir()
    if filename and '/' in filename:
        raise RuntimeError('Unexpected slash in filename. You should consider setting `--doc-sub-dir` instead')
    path_tag_summary = pth_docs / (filename or from_ctx(ctx, 'tags', 'filename'))
    patterns = (ignore_patterns or from_ctx(ctx, 'tags', 'ignore_patterns')).split(',')
    paths_source = find_project_files(pth_base_dir, ignore_patterns=[*filter(lambda item: item, patterns)])

    write_code_tag_file(
        path_tag_summary=path_tag_summary,
        paths_source=paths_source,
        base_dir=pth_base_dir,
        regex=regex,
        tags=tag_order,
        header='# Collected Code Tags',
    )



"""Support can-skip logic from Make."""

from pathlib import Path

from beartype import beartype
from beartype.typing import List
from corallium.log import logger


@beartype
def can_skip(*, prerequisites: List[Path], targets: List[Path]) -> bool:
    """Generic make-style task skipping logic based on file `mtime`.

    Example use with Invoke, but can be used anywhere:

    ```py
    @task
    def test(ctx: Context) -> None:
        if can_skip(prerequisites=[*Path('src').rglob('*.py')], targets=[Path('.coverage.xml')]):
            return  # Exit early

        ...  # Task code
    ```

    """
    if not (ts_prerequisites := [pth.stat().st_mtime for pth in prerequisites]):
        raise ValueError('Required files do not exist', prerequisites)

    ts_targets = [pth.stat().st_mtime for pth in targets]
    if ts_targets and min(ts_targets) > max(ts_prerequisites):
        logger.warning('Skipping because targets are newer', targets=targets)
        return True
    return False


@beartype
def dont_skip(*, prerequisites: List[Path], targets: List[Path]) -> bool:
    """To use for testing with mock; always returns False."""
    logger.debug('Mocking can_skip', prerequisites=prerequisites, targets=targets)
    return False



"""Collect code tags and output for review in a single location."""

import re
from collections import defaultdict
from contextlib import suppress
from functools import lru_cache
from pathlib import Path
from subprocess import CalledProcessError  # nosec  # noqa: S404

import arrow
import pandas as pd
from beartype import beartype
from beartype.typing import Dict, List, Pattern, Sequence, Tuple
from corallium.file_helpers import read_lines
from corallium.log import logger
from corallium.shell import capture_shell
from pydantic import BaseModel, ConfigDict

SKIP_PHRASE = 'calcipy_skip_tags'
"""String that indicates the file should be excluded from the tag search."""

COMMON_CODE_TAGS = ['FIXME', 'TODO', 'PLANNED', 'HACK', 'REVIEW', 'TBD', 'DEBUG']
"""Most common code tags.

FYI and NOTE are excluded to not be tracked in the Code Summary.

"""

CODE_TAG_RE = r'((^|\s|\(|"|\')(?P<tag>{tag})(:| -)([^\r\n]))(?P<text>.+)'
"""Default code tag regex with `tag` and `text` matching groups.

Requires formatting with list of tags: `CODE_TAG_RE.format(tag='|'.join(tag_list))`

Commonly, the `tag_list` could be `COMMON_CODE_TAGS`

"""


class _CodeTag(BaseModel):
    """Code Tag (FIXME,TODO,etc) with contextual information."""

    lineno: int
    tag: str
    text: str
    model_config = ConfigDict(frozen=True)


class _Tags(BaseModel):
    """Collection of code tags with additional contextual information."""

    path_source: Path
    code_tags: List[_CodeTag]
    model_config = ConfigDict(frozen=True)


@beartype
def _search_lines(
    lines: List[str], regex_compiled: Pattern[str],
    skip_phrase: str = 'calcipy_skip_tags',
) -> List[_CodeTag]:
    """Search lines of text for matches to the compiled regular expression.

    Args:
        lines: lines of text as list
        regex_compiled: compiled regular expression. Expected to have matching groups `(tag, text)`
        skip_phrase: skip file if string is found in final two lines. Default is `SKIP_PHRASE`

    Returns:
        List[_CodeTag]: list of all code tags found in lines

    """
    if skip_phrase in '\n'.join(lines[-2:]):
        return []

    max_len = 400
    comments = []
    for lineno, line in enumerate(lines):
        if match := regex_compiled.search(line):
            if len(line) <= max_len:  # FYI: Suppress long lines
                group = match.groupdict()
                comments.append(_CodeTag(lineno=lineno + 1, tag=group['tag'], text=group['text']))
            else:
                logger.text_debug('Skipping long line', lineno=lineno, line=line[:200])
    return comments


@beartype
def _search_files(paths_source: Sequence[Path], regex_compiled: Pattern[str]) -> List[_Tags]:
    """Collect matches from multiple files.

    Args:
        paths_source: list of source files to parse
        regex_compiled: compiled regular expression. Expected to have matching groups `(tag, text)`

    Returns:
        List[_Tags]: list of all code tags found in files

    """
    matches = []
    for path_source in paths_source:
        lines = []
        try:
            lines = read_lines(path_source)
        except UnicodeDecodeError as err:
            logger.text_debug('Could not parse', path_source=path_source, err=err)

        if comments := _search_lines(lines, regex_compiled):
            matches.append(_Tags(path_source=path_source, code_tags=comments))

    return matches


_GITHUB_ORIGIN = r'^.+github.com[:/](?P<owner>[^/]+)/(?P<repository>[^.]+)(?:\.git)?$'
"""Match owner and repository from a GitHub git origin URI."""


@beartype
def github_blame_url(clone_uri: str) -> str:
    """Format the blame URL.

    Args:
        clone_uri: git remote URI

    Returns:
       str: `repo_url`

    """
    # Could be ssh or http (with or without .git)
    # > git@github.com:KyleKing/calcipy.git
    # > https://github.com/KyleKing/calcipy.git
    if matches := re.compile(_GITHUB_ORIGIN).match(clone_uri):
        github_url = 'https://github.com/'
        return f"{github_url}{matches['owner']}/{matches['repository']}"
    return ''


@lru_cache(maxsize=128)
def _git_info(cwd: Path) -> Tuple[Path, str]:
    """Collect information about the local git repository.

    Based on snippets from: https://gist.github.com/abackstrom/4034721#gistcomment-3982270
    and: https://github.com/rscherf/GitLink/blob/e2e7c412630246efc86de4fe71192f15bf11209e/GitLink.py

    Args:
        cwd: Path to the current working directory (typically file_path.parent)

    Returns:
        Tuple[Path, str]: (git_dir, repo_url)

    """
    git_dir = Path(capture_shell('git rev-parse --show-toplevel', cwd=cwd))
    clone_uri = ''
    with suppress(CalledProcessError):
        clone_uri = capture_shell('git remote get-url origin', cwd=cwd)
    return git_dir, github_blame_url(clone_uri)


class _CollectorRow(BaseModel):
    """Each row of the Code Tag table."""

    tag_name: str
    comment: str
    last_edit: str
    source_file: str

    @classmethod
    @beartype
    def from_code_tag(cls, code_tag: _CodeTag, last_edit: str, source_file: str) -> '_CollectorRow':
        return cls(
            tag_name=f'{code_tag.tag:>7}',
            comment=code_tag.text,
            last_edit=last_edit,
            source_file=source_file,
        )


@beartype
def _format_from_blame(
    *, collector_row: _CollectorRow, blame: str, repo_url: str, cwd: Path, rel_path: Path,
) -> _CollectorRow:
    """Parse the git blame for useful timestamps and author when available."""
    # Note: line number may be different in older blame (and relative path)
    revision, old_line_number = blame.split('\n')[0].split(' ')[:2]
    # If the change has not yet been committed, use the branch name as best guess
    if all(_c == '0' for _c in revision):
        revision = capture_shell('git branch --show-current', cwd=cwd)
    # Format a nice timestamp of the last edit to the line
    blame_dict = {
        line.split(' ')[0]: ' '.join(line.split(' ')[1:])
        for line in blame.split('\n')
    }

    # Handle uncommitted files that only have author-time and author-tz
    user = 'committer' if 'committer-tz' in blame_dict else 'author'
    dt = arrow.get(int(blame_dict[f'{user}-time']))
    tz = blame_dict[f'{user}-tz'][:3] + ':' + blame_dict[f'{user}-tz'][-2:]
    collector_row.last_edit = arrow.get(dt.isoformat()[:-6] + tz).format('YYYY-MM-DD')

    if repo_url:
        # Filename may not be present if uncommitted. Use local path as fallback
        remote_file_path = blame_dict.get('filename', rel_path.as_posix())
        # Assumes Github format
        git_url = f'{repo_url}/blame/{revision}/{remote_file_path}#L{old_line_number}'
        collector_row.source_file = f'[{collector_row.source_file}]({git_url})'

    return collector_row


@beartype
def _format_record(base_dir: Path, file_path: Path, comment: _CodeTag) -> _CollectorRow:
    """Format each table row for the code tag summary file. Include git permalink.

    Args:
        base_dir: base path of the project if git directory is not known
        file_path: path to the file of interest
        comment: _CodeTag information for the matched tag

    Returns:
        Dict[str, str]: formatted dictionary with file info

    """
    cwd = file_path.parent
    _git_dir, repo_url = _git_info(cwd=cwd)

    # Set fallback values if git logic doesn't work
    rel_path = file_path.relative_to(base_dir)
    collector_row = _CollectorRow.from_code_tag(
        code_tag=comment,
        last_edit='N/A',
        source_file=f'{rel_path.as_posix()}:{comment.lineno}',
    )

    try:
        blame = capture_shell(f'git blame {file_path} -L {comment.lineno},{comment.lineno} --porcelain', cwd=cwd)
        collector_row = _format_from_blame(
            collector_row=collector_row, blame=blame, repo_url=repo_url, cwd=cwd, rel_path=rel_path,
        )
    except CalledProcessError as exc:
        handled_errors = (128,)
        if exc.returncode not in handled_errors:
            raise
        logger.text_debug('Skipping blame', file_path=file_path, exc=exc)

    return collector_row


@beartype
def _format_report(  # noqa: CAC001
    base_dir: Path, code_tags: List[_Tags], tag_order: List[str],
) -> str:
    """Pretty-format the code tags by file and line number.

    Args:
        base_dir: base directory relative to the searched files
        code_tags: list of all code tags found in files
        tag_order: subset of all tags to include in the report and specified order

    Returns:
        str: pretty-formatted text

    """
    output = ''
    records = []
    counter: Dict[str, int] = defaultdict(lambda: 0)
    for comments in sorted(code_tags, key=lambda tc: tc.path_source, reverse=False):
        for comment in comments.code_tags:
            if comment.tag in tag_order:
                collector_row = _format_record(base_dir, comments.path_source, comment)
                records.append({
                    'Type': collector_row.tag_name,
                    'Comment': collector_row.comment,
                    'Last Edit': collector_row.last_edit,
                    'Source File': collector_row.source_file,
                })
                counter[comment.tag] += 1
    if records:
        df_tags = pd.DataFrame(records)
        # Prevent URLs from appearing on multiple lines
        content = df_tags.to_markdown(index=False, tablefmt='github', maxcolwidths=None) or ''
        for line in content.split('\n'):
            if not line.startswith('/'):
                output += '\n'
            output += line
    logger.text_debug('counter', counter=counter)

    sorted_counter = {tag: counter[tag] for tag in tag_order if tag in counter}
    logger.text_debug('sorted_counter', sorted_counter=sorted_counter)
    if formatted_summary := ', '.join(
        f'{tag} ({count})' for tag, count in sorted_counter.items()
    ):
        output += f'\n\nFound code tags for {formatted_summary}\n'
    return output


@beartype
def write_code_tag_file(
    path_tag_summary: Path,
    paths_source: List[Path],
    base_dir: Path,
    regex: str = '',
    tags: str = '',
    header: str = '# Task Summary\n\nAuto-Generated by `calcipy`',
) -> None:
    """Create the code tag summary file.

    Args:
        path_tag_summary: Path to the output file
        paths_source: list of source files to parse
        base_dir: base directory relative to the searched files
        regex: compiled regular expression. Expected to have matching groups `(tag, text)`.
            Default is CODE_TAG_RE with tags from tag_order
        tags: subset of all tags to include in the report and specified order. Default is COMMON_CODE_TAGS
        header: header text

    """
    tag_order = [_t.strip() for _t in tags.split(',') if _t] or COMMON_CODE_TAGS
    matcher = (regex or CODE_TAG_RE).format(tag='|'.join(tag_order))

    matches = _search_files(paths_source, re.compile(matcher))
    if report := _format_report(
        base_dir, matches, tag_order=tag_order,
    ).strip():
        path_tag_summary.parent.mkdir(exist_ok=True, parents=True)
        path_tag_summary.write_text(f'{header}\n\n{report}\n\n<!-- {SKIP_PHRASE} -->\n')
        logger.text('Created Code Tag Summary', path_tag_summary=path_tag_summary)
    elif path_tag_summary.is_file():
        path_tag_summary.unlink()



"""Check for stale packages."""

import asyncio
import json
from functools import lru_cache
from pathlib import Path

import arrow
import httpx
import numpy as np
from arrow import Arrow
from beartype import beartype
from beartype.typing import Dict, List, Optional, Union
from bidict import bidict
from corallium.file_helpers import LOCK
from corallium.log import logger
from corallium.tomllib import tomllib
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator
from pyrate_limiter import Rate
from pyrate_limiter.limiter import Limiter

from .. import can_skip  # Required for mocking can_skip.can_skip

CALCIPY_CACHE = Path('.calcipy_packaging.lock')
"""Path to the packaging lock file."""


class _HostedPythonPackage(BaseModel):
    """Representative information for a python package hosted on some domain."""

    # Note: "releases" was removed from the versioned URL: https://warehouse.pypa.io/api-reference/json.html#release
    domain: str = Field(default='https://pypi.org/pypi/{name}/json')
    name: str
    version: str
    datetime: Optional[Arrow] = Field(default=None)
    latest_version: str = Field(default='')
    latest_datetime: Optional[Arrow] = Field(default=None)
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer('datetime', 'latest_datetime')
    def serialize_datetime(self, value: Optional[Arrow]) -> Optional[str]:  # noqa: PLR6301
        return str(value) if value else None

    @field_validator('datetime', 'latest_datetime', mode='before')
    @classmethod
    def date_validator(cls, value: Union[str, Arrow]) -> Arrow:
        return arrow.get(value)


@lru_cache(maxsize=1)
def _limiter() -> Limiter:
    """Configure rate-limiter.

    https://pypi.org/project/pyrate-limiter

    """
    rate = Rate(limit=3, interval=5)  # Duration.seconds
    return Limiter(rate, max_delay=600)


async def _get_release_date(package: _HostedPythonPackage) -> _HostedPythonPackage:
    """Retrieve release date metadata for the specified package.

    Args:
        package: `_HostedPythonPackage`

    Returns:
        _HostedPythonPackage: updated with release date metadata from API

    """
    _limiter().try_acquire('pypi')

    # Retrieve the JSON summary for the specified package
    json_url = package.domain.format(name=package.name)
    async with httpx.AsyncClient() as client:
        res = await client.get(json_url, timeout=30)   # nosem
        res.raise_for_status()
        res_json = res.json()
        if not (releases := res_json['releases']):
            msg = f'Failed to locate "releases" or "urls" from {json_url}: {res_json}'
            raise RuntimeError(msg)

    # Also retrieve the latest release date of the package looking through all releases
    release_dates = bidict({
        arrow.get(release_data[0]['upload_time_iso_8601']): version
        for version, release_data in releases.items()
        if release_data
    })
    try:
        package.datetime = release_dates.inverse[package.version]
    except KeyError:  # pragma: no cover
        msg = f'Could not locate {package} in {res_json}. Please wait and try again later'
        raise RuntimeError(msg) from None
    package.latest_datetime = max([*release_dates])
    package.latest_version = release_dates[package.latest_datetime]
    return package


@beartype
def _read_cache(path_pack_lock: Path = CALCIPY_CACHE) -> Dict[str, _HostedPythonPackage]:
    """Read the cached packaging information.

    Args:
        path_pack_lock: Path to the lock file. Default is `CALCIPY_CACHE`

    Returns:
        Dict[str, _HostedPythonPackage]: the cached packages

    """
    if not path_pack_lock.is_file():
        path_pack_lock.write_text('{}', encoding='utf-8')  # noqa: P103
    old_cache: Dict[str, Dict[str, str]] = json.loads(path_pack_lock.read_text(encoding='utf-8'))
    return {
        package_name: _HostedPythonPackage(**meta_data)  # type: ignore[arg-type]
        for package_name, meta_data in old_cache.items()
    }


@beartype
def _collect_release_dates(
    packages: List[_HostedPythonPackage],
    old_cache: Optional[Dict[str, _HostedPythonPackage]] = None,
) -> List[_HostedPythonPackage]:
    """Use the cache to retrieve only metadata that needs to be updated.

    Args:
        packages: list of `_HostedPythonPackage`
        old_cache: cache data to compare against to limit network requests

    Returns:
        List[_HostedPythonPackage]: packages with updated release dates

    """
    if old_cache is None:
        old_cache = {}

    updated_packages = []
    for package in packages:
        try:
            cached_package = old_cache.get(package.name)
            cached_version = '' if cached_package is None else cached_package.version
            if package.version != cached_version:
                updated_packages.append(asyncio.run(_get_release_date(package)))
            elif cached_package:
                updated_packages.append(cached_package)
        except httpx.HTTPError as exc:  # noqa: PERF203
            logger.warning('Could not lock package', package=package, error=str(exc))
    return updated_packages


@beartype
def _write_cache(updated_packages: List[_HostedPythonPackage], path_pack_lock: Path = CALCIPY_CACHE) -> None:
    """Read the cached packaging information.

    Args:
        updated_packages: updated packages to store
        path_pack_lock: Path to the lock file. Default is `CALCIPY_CACHE`

    """
    new_cache = {pack.name: json.loads(pack.model_dump_json()) for pack in updated_packages}
    pretty_json = json.dumps(new_cache, indent=4, separators=(',', ': '), sort_keys=True)
    path_pack_lock.write_text(pretty_json + '\n', encoding='utf-8')


@beartype
def _read_packages(path_lock: Path) -> List[_HostedPythonPackage]:
    """Read packages from lock file. Currently only support `poetry.lock`, but could support more in the future.

    Args:
        path_lock: Path to the lock file to parse

    Returns:
        List[_HostedPythonPackage]: packages found in the lock file

    Raises:
        NotImplementedError: if a lock file other that the poetry lock file is used

    """
    if path_lock.name != LOCK.name:
        msg = f'"{path_lock.name}" is not a currently supported lock type. Try "{LOCK.name}" instead'
        raise NotImplementedError(msg)

    lock = tomllib.loads(path_lock.read_text(errors='ignore'))
    return [
        _HostedPythonPackage(
            name=dependency['name'], version=dependency['version'],
        ) for dependency in lock['package']
    ]


@beartype
def _packages_are_stale(packages: List[_HostedPythonPackage], *, stale_months: int) -> bool:
    """Check for stale packages. Raise error and log all stale versions found.

    Args:
        packages: List of packages
        stale_months: cutoff in months for when a package might be stale enough to be a risk

    """
    @beartype
    def format_package(pack: _HostedPythonPackage) -> str:
        delta = pack.datetime.humanize()  # type: ignore[union-attr]
        latest = '' if pack.version == pack.latest_version else f' (*New version available: {pack.latest_version}*)'
        return f'- {delta}: {pack.name} {pack.version}{latest}'

    now = arrow.utcnow()
    stale_cutoff = now.shift(months=-1 * stale_months)
    stale_packages = [pack for pack in packages if not pack.datetime or pack.datetime < stale_cutoff]
    # TODO: If no stale, write out five oldest?
    if stale_packages:
        pkgs = sorted(stale_packages, key=lambda x: x.datetime or stale_cutoff)
        stale_list = '\n'.join([format_package(_p) for _p in pkgs])
        logger.warning('Found stale packages that may be a dependency risk', stale_list=stale_list)
        return True
    if packages:
        oldest_date = np.amin([pack.datetime for pack in packages])  # pyright: ignore[reportGeneralTypeIssues]
        logger.text('No stale packages found', oldest=oldest_date.humanize(), stale_threshold=stale_months)
    return False


@beartype
def check_for_stale_packages(*, stale_months: int, path_lock: Path = LOCK, path_cache: Path = CALCIPY_CACHE) -> bool:
    """Read the cached packaging information.

    Args:
        stale_months: cutoff in months for when a package might be stale enough to be a risk

    """
    packages = _read_packages(path_lock)
    cached_packages = _read_cache(path_cache)
    if cached_packages and can_skip.can_skip(prerequisites=[path_lock], targets=[path_cache]):
        packages = [*cached_packages.values()]
    else:
        packages = _collect_release_dates(packages, cached_packages)
        _write_cache(packages, path_cache)
    return _packages_are_stale(packages, stale_months=stale_months)



"""Calcipy-Invoke Defaults."""

import json
from contextlib import suppress
from pathlib import Path

from beartype import beartype
from invoke.context import Context

from ._invoke import Collection

DEFAULTS = {
    'tags': {
        'filename': 'CODE_TAG_SUMMARY.md',
        'ignore_patterns': '',
    },
    'test': {
        'min_cover': '0',
        'out_dir': 'releases/tests',
    },
    'type': {
        'out_dir': 'releases/tests/mypy_html',
    },
}


@beartype
def from_ctx(ctx: Context, group: str, key: str) -> str:
    """Safely extract the value from the context or the defaults.

    Instead of `ctx.tests.out_dir` use `from_ctx(ctx, 'test', 'out_dir')`

    """
    with suppress(KeyError):
        return str(ctx.config[group][key])
    return str(DEFAULTS[group][key])


@beartype
def new_collection() -> Collection:
    """Initialize a collection with the combination of merged and project-specific defaults."""
    ns = Collection('')

    # Merge default and user configuration
    ns.configure(DEFAULTS)
    config_path = Path('.calcipy.json')
    if config_path.is_file():
        ns.configure(json.loads(config_path.read_text(encoding='utf-8')))

    return ns



"""Extend Invoke for Calcipy."""

import logging
import os
from functools import wraps
from pathlib import Path
from types import ModuleType

from beartype import beartype
from beartype.typing import Any, Callable, Dict, List, Optional, Tuple, Union
from corallium.log import configure_logger, logger
from invoke.collection import Collection as InvokeCollection  # noqa: TID251
from invoke.context import Context
from invoke.tasks import Task
from pydantic import BaseModel, Field, PositiveInt

TASK_ARGS_ATTR = 'dev_args'
TASK_KWARGS_ATTR = 'dev_kwargs'

DeferedTask = Union[Callable, Task]  # type: ignore[type-arg]


class GlobalTaskOptions(BaseModel):
    """Global Task Options."""

    working_dir: Path = Field(default_factory=Path.cwd)
    """Working directory for the program to use globally."""

    file_args: List[Path] = Field(default_factory=list)
    """List of Paths to modify."""

    verbose: PositiveInt = Field(default=0, le=3)
    """Verbosity level."""

    keep_going: bool = False
    """Continue task execution regardless of failure."""


@beartype
def _configure_task_logger(ctx: Context) -> None:  # pragma: no cover
    """Configure the logger based on task context."""
    verbose = ctx.config.gto.verbose
    log_lookup = {3: logging.NOTSET, 2: logging.DEBUG, 1: logging.INFO, 0: logging.WARNING}
    raw_log_level = log_lookup.get(verbose)
    log_level = logging.ERROR if raw_log_level is None else raw_log_level
    configure_logger(log_level=log_level)


@beartype
def _run_task(func: Any, ctx: Context, *args: Any, show_task_info: bool, **kwargs: Any) -> Any:  # pragma: no cover
    """Run the task function with optional logging."""
    if show_task_info:
        summary = func.__doc__.split('\n')[0]
        logger.text(f'Running {func.__name__}', is_header=True, summary=summary)
        logger.text_debug('With task arguments', args=args, kwargs=kwargs)

    result = func(ctx, *args, **kwargs)

    if show_task_info:
        logger.text('')
        logger.text_debug(f'Completed {func.__name__}', result=result)

    return result


@beartype
def _wrapped_task(ctx: Context, *args: Any, func: Any, show_task_info: bool, **kwargs: Any) -> Any:  # pragma: no cover
    """Extended task logic."""
    try:
        ctx.config.gto  # noqa: B018
    except AttributeError:
        ctx.config.gto = GlobalTaskOptions()

    # Begin utilizing Global Task Options
    os.chdir(ctx.config.gto.working_dir)
    _configure_task_logger(ctx)

    try:
        return _run_task(func, ctx, *args, show_task_info=show_task_info, **kwargs)
    except Exception:
        if not ctx.config.gto.keep_going:
            raise
        logger.exception('Task Failed', func=str(func), args=args, kwargs=kwargs)
    return None


def _build_task(task: DeferedTask) -> Task:  # type: ignore[type-arg]  # pragma: no cover
    """Defer creation of the Task."""
    if hasattr(task, TASK_ARGS_ATTR) or hasattr(task, TASK_KWARGS_ATTR):
        @wraps(task)
        def inner(*args: Any, **kwargs: Any) -> Any:
            return _wrapped_task(*args, func=task, show_task_info=show_task_info, **kwargs)

        kwargs = getattr(task, TASK_KWARGS_ATTR)
        show_task_info = kwargs.pop('show_task_info', None) or False
        pre = [_build_task(pre) for pre in kwargs.pop('pre', None) or []]
        post = [_build_task(post) for post in kwargs.pop('post', None) or []]
        return Task(inner, *getattr(task, TASK_ARGS_ATTR), pre=pre, post=post, **kwargs)  # type: ignore[misc,arg-type]
    return task  # type: ignore[return-value]


class Collection(InvokeCollection):

    @classmethod
    def from_module(
        cls,
        module: ModuleType,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        loaded_from: Optional[str] = None,
        auto_dash_names: Optional[bool] = None,
    ) -> 'InvokeCollection':
        """Extend search for a namespace, Task, or deferred task."""
        collection = super().from_module(
            module=module,
            name=name,
            config=config,
            loaded_from=loaded_from,
            auto_dash_names=auto_dash_names,
        )

        # If tasks were not loaded from a namespace or otherwise found
        if not collection.task_names:
            # Look for any decorated, but deferred "Tasks"
            for task in (fxn for fxn in vars(module).values() if hasattr(fxn, TASK_ARGS_ATTR)):
                collection.add_task(task)

        return collection

    def add_task(
        self,
        task: DeferedTask,
        name: Optional[str] = None,
        aliases: Optional[Tuple[str, ...]] = None,
        default: Optional[bool] = None,
    ) -> None:
        """Extend for deferred tasks."""
        super().add_task(task=_build_task(task), name=name, aliases=aliases, default=default)



"""Experiment with bumping the git tag using `griffe`."""

import griffe
import semver
from beartype import beartype
from corallium.log import logger
from griffe.exceptions import BuiltinModuleError


@beartype
def bump_tag(*, pkg_name: str, tag: str, tag_prefix: str) -> str:
    """Make a SemVer minor bump using `griffe` if there were any breaking changes.

    Major versions must be bumped manually

    """
    previous = griffe.load_git(pkg_name, ref=tag)
    current = griffe.load(pkg_name)

    breakages = [*griffe.find_breaking_changes(previous, current)]
    for breakage in breakages:
        try:
            logger.text(breakage._explain_oneline())  # noqa: SLF001
        except BuiltinModuleError:  # noqa: PERF203
            logger.warning(str(breakage))
        except Exception:
            logger.exception(str(breakage))

    try:
        ver = semver.Version.parse(tag.replace(tag_prefix, ''))
    except ValueError:
        logger.exception('Failed to parse tag', tag=tag)
        return ''
    new_ver = ver.bump_minor() if any(breakages) else ver.bump_patch()
    return f'{tag_prefix}{new_ver}'



"""Document CLI."""

import webbrowser
from contextlib import suppress
from pathlib import Path

from beartype import beartype
from corallium.file_helpers import (
    MKDOCS_CONFIG,
    delete_dir,
    ensure_dir,
    open_in_browser,
    read_package_name,
    read_yaml_file,
    trim_trailing_whitespace,
)
from invoke.context import Context
from invoke.exceptions import UnexpectedExit

from ..cli import task
from ..invoke_helpers import get_doc_subdir, get_project_path, run
from ..md_writer import write_autoformatted_md_sections
from .executable_utils import python_dir


@beartype
def _diagram_task(ctx: Context, pdoc_out_path: Path) -> None:
    """Return actions to generate code diagrams in the module documentation directory.

    Note: must be run after `document` because pdoc will delete these files

    PUML support may be coming in a future release: https://github.com/PyCQA/pylint/issues/4498

    Args:
        pdoc_out_path: path to the top-level pdoc output. Expect subdir with module name

    """
    diagram_md = """# Code Diagrams

> Auto-generated with `pylint-pyreverse`

## Packages

![./packages.svg](./packages.svg)

[Full Size](./packages.svg)

## Classes

![./classes.svg](./classes.svg)

[Full Size](./classes.svg)
"""
    pkg_name = read_package_name()
    path_diagram = pdoc_out_path / pkg_name / '_code_diagrams.md'
    ensure_dir(path_diagram.parent)
    path_diagram.write_text(diagram_md)
    run(ctx, f'{python_dir()}/pyreverse {pkg_name} --output svg --output-directory {path_diagram.parent}')


@beartype
def get_out_dir() -> Path:
    """Retrieve the mkdocs-specified site directory."""
    mkdocs_config = read_yaml_file(get_project_path() / MKDOCS_CONFIG)
    return Path(mkdocs_config.get('site_dir', 'releases/site'))


@task()
def build(ctx: Context) -> None:
    """Build documentation with mkdocs."""
    auto_doc_path = get_doc_subdir().parent / 'modules'
    write_autoformatted_md_sections()
    delete_dir(auto_doc_path)
    _diagram_task(ctx, auto_doc_path)

    # Find and trim trailing whitespace
    for path_md in auto_doc_path.rglob('*.md'):
        trim_trailing_whitespace(path_md)

    run(ctx, f'{python_dir()}/mkdocs build --site-dir {get_out_dir()}')


@beartype
def _is_mkdocs_local() -> bool:
    """Check if mkdocs is configured for local output.

    See notes on local-link configuration here: https://github.com/timothycrosley/portray/issues/65

    Additional information on using local search here: https://github.com/wilhelmer/mkdocs-localsearch

    Returns:
        bool: True if configured for local file output rather than hosted

    """
    mkdocs_config = read_yaml_file(get_project_path() / MKDOCS_CONFIG)
    return mkdocs_config.get('use_directory_urls') is False


@task()
def watch(ctx: Context) -> None:
    """Serve local documentation for local editing."""
    if _is_mkdocs_local():  # pragma: no cover
        path_doc_index = get_out_dir() / 'index.html'
        open_in_browser(path_doc_index)
    else:  # pragma: no cover
        webbrowser.open('http://localhost:8000')
        run(ctx, f'{python_dir()}/mkdocs serve --dirtyreload')


@task()
def deploy(ctx: Context) -> None:
    """Deploy docs to the Github `gh-pages` branch."""
    if _is_mkdocs_local():  # pragma: no cover
        raise NotImplementedError('Not yet configured to deploy documentation without "use_directory_urls"')

    with suppress(UnexpectedExit):
        run(ctx, 'pre-commit uninstall')  # To prevent pre-commit failures when mkdocs calls push
    run(ctx, f'{python_dir()}/mkdocs gh-deploy --force')
    with suppress(UnexpectedExit):
        run(ctx, 'pre-commit install')  # Restore pre-commit
